#!/usr/bin/env python
# -*- coding: utf-8 -*-

__all__ = [
    'abspath',
    'clean_parallel',
    'exec_parallel',
    'guess_langcode',
    'progress',
    'prompt',
    'random_split',
    'wait_files',
]




#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''Pipe-view script for I/O progress'''

# Standard libraries
import argparse
import sys

# Local libraries
from lpu.common import logging
from lpu.common.progress import pipe_view

logger = logging.getColorLogger(__name__)

def cmdPipeView(args):
    REFRESH_INTERVAL = 1
    parser = argparse.ArgumentParser(description='Show the progress of pipe I/O')
    parser.add_argument('filepaths', metavar="filepath", nargs="*", type=str, help='path of file to load')
    parser.add_argument('--lines', '-l', action='store_true', help='line count mode (default: byte count mode)')
    parser.add_argument('--refresh', '-r', type=float, default=1.0, help='refresh interval (default: %(default)s')
    parser.add_argument('--header', '-H', type=str, help='header of the progress information')
    parsed = parser.parse_args(args)
    logger.debug(parsed)
    mode = 'bytes'
    if parsed.lines:
        mode = 'lines'
    pipe_view(parsed.filepaths, mode=mode, header=parsed.header, refresh=parsed.refresh)

def main():
    cmdPipeView(sys.argv[1:])

if __name__ == '__main__':
    main()




# distutils: language=c++
# -*- coding: utf-8 -*-

'''tree expression and operations'''

import copy
from collections import Iterable

# third party library
#import numpy as np

cdef class TreeNode:
    cdef readonly list children
    #cdef Forest children
    #cdef object label
    cdef public str label

    #def __cinit__(self, object label):
    def __cinit__(self, str label):
        self.children = []
        #self.children = Forest()
        self.label = label

    cpdef TreeNode append(self, node):
        if isinstance(node, TreeNode):
            self.children.append(node)
        elif isinstance(node, str):
            self.children.append(TreeNode(node))
        else:
            raise TypeError('expected TreeNode or str, given %s' % type(node))
        return self

    cpdef checkValid(self, deep=True):
        cdef TreeNode node
        if len(self.label) == 0 and len(self.children) > 0:
            return False
        if not deep:
            return True
        for node in self.children:
#            if not node.checkValid(True):
                return False
        return True

    @staticmethod
    def fromS(str expr):
        cdef TreeNode t = TreeNode('')
        cdef TreeNode tmp
        return t

    cpdef str toStr(self):
#        cdef str strChildren = ''
#        map(Tree.toStr, self.children)
        cdef str strChildren = ''
        cdef TreeNode node
        if not self.checkValid(False):
            raise ValueError( (self.label, self.children) )
        if self.children:
            for node in self.children:
                if strChildren:
                    strChildren += ' '
                strChildren += node.toStr()
            return '(' + self.label + ' ' + strChildren + ')'
        else:
            return '(' + self.label + ')'

    def __str__(self):
        return self.toStr()
    def __repr__(self):
        cdef str mod = str(self.__class__.__module__)
        try:
            return "%s.fromS(%r)" % (mod,self.toStr())
        except:
            return "%s.fromS(__invalid__)" % (mod,)

Tree = TreeNode

#cdef __findMin(str target, sub, start = 0, end = None):
#    cdef str key
#    cdef int found
#    cdef int minFound = -1
#    if isinstance(sub, str):
#        return target.find(sub, start, end)
#    elif isinstance(sub, Iterable):
#        for key in sub:
#            found = target.find(key, start, end)
#            if minFound < 0:
#                minFound = found
#            else:
#                minFound = min(minFound, found)
#        return minFound

cpdef object parseSExpression(str expr, int i = 0):
    cdef object cont = ''
    cdef object item
    while i < len(expr):
        #dprint('Expr[%s]: %s' % (i, expr[i]))
        if expr[i] == '(':
            #dprint('Push')
            cont = []
            while i < len(expr):
                if expr[i] == ')':
                    #dprint("Closing: %s" % cont)
                    return cont, i + 1
                item, i = parseSExpression(expr, i + 1)
                if item:
                    #print("Appending: %s" % item)
                    cont.append(item)
            return cont, i
        elif expr[i] == ')':
            #dprint("Closing: " + cont)
            return cont, i
        elif expr[i] == ' ':
            #dprint('Elem: ' + cont)
            return cont, i
        else:
            cont += expr[i]
            i += 1
    return cont, i

def indexTree(list tree):
    #cdef list indexToTree = []
    cdef list indexToLabel = []
    cdef list indexToLeftRange = []
    def appendNodePostOrder(object node):
        cdef leftRange = -1
        cdef index
        if isinstance(node, list):
            if len(node) > 0:
                # top/inner node
                for sub in node[1:]:
                    # child node
                    #appendNodePostOrder(sub)
                    index = appendNodePostOrder(sub)
                    if leftRange < 0 or index < leftRange:
                        leftRange = index
                #indexToTree.append(node)
                indexToLabel.append(node[0])
                indexToLeftRange.append(leftRange)
                return leftRange
        else:
            # leaf
            leftRange = len(indexToLabel)
            indexToLabel.append(node)
            indexToLeftRange.append(leftRange)
            #return len(indexToLabel) - 1
            return leftRange
    appendNodePostOrder(tree)
    #return indexToTree
    return indexToLabel, indexToLeftRange

def countElements(list tree):
    def innerCount(object node):
        cdef int numElems = 0
        if isinstance(node, list):
            if len(node) > 0:
                for sub in node[1:]:
                    numElems += innerCount(sub)
            return numElems + 1
        else:
            # leaf
            return 1
    return innerCount(tree)

#import pprint
cpdef calcEditDistance(seq1, seq2):
    cdef list memo = [[None]*(len(seq2)+1) for _ in range(0,len(seq1)+1)]
    cdef int i, j
    for i in range(0, len(seq1)+1):
        memo[i][0] = i
    for j in range(0, len(seq2)+1):
        memo[0][j] = j
    for i in range(1, len(seq1)+1):
        for j in range(1, len(seq2)+1):
            #print("--")
            #print("i: %s (%s), j: %s (%s)" % (i, seq1[:i], j, seq2[:j]))
            if seq1[i-1] == seq2[j-1]:
                memo[i][j] = min(memo[i-1][j]+1, memo[i][j-1]+1, memo[i-1][j-1])
            else:
                memo[i][j] = min(memo[i-1][j]+1, memo[i][j-1]+1, memo[i-1][j-1]+1)
            #pprint.pprint(memo)
    #pprint.pprint(memo)
    return memo[len(seq1)][len(seq2)]

import pprint
def calcTreeEditDistance(object tree1, object tree2):
    cdef dict memo = {}
    #cdef object memo
    cdef list indexToLabel1
    cdef list indexToLeftRange1
    cdef list indexToLabel2
    cdef list indexToLeftRange2
    cdef int numNodes1
    cdef int numNodes2
    if isinstance(tree1, str) and tree1.strip()[0] == '(':
        # S-expression
        tree1 = parseSExpression(tree1)[0]
    if isinstance(tree2, str) and tree2.strip()[0] == '(':
        # S-expression
        tree2 = parseSExpression(tree2)[0]
    indexToLabel1, indexToLeftRange1 = indexTree(tree1)
    indexToLabel2, indexToLeftRange2 = indexTree(tree2)
    numNodes1 = len(indexToLabel1)
    numNodes2 = len(indexToLabel2)
    #memo = - np.ones([numNodes1,numNodes1,numNodes2,numNodes2])
    ## Forest2 Right
    #memo = [None] * for len(indexToLabel2)
    ## Forest2 Left
    #memo = [copy.deepcopy(memo) for _ in indexToLabel2]
    ## Forest1 Right
    #memo = [copy.deepcopy(memo) for _ in indexToLabel1]
    ## Forest1 Left
    #memo = [copy.deepcopy(memo) for _ in indexToLabel1]
    def calcInnerDistance(int left1, int right1, int left2, int right2):
        #print("(%s,%s) vs (%s,%s)" % (left1, right1, left2, right2))
        cdef v1, v2, v3, v
        if left1 < 0 or left2 < 0:
            return 0
        if right1 < left1:
            # empty
            return max(0, right2 - left2 + 1)
        if right2 < left2:
            # empty
            return max(0, right1 - left1 + 1)
        #if left1 < 0 or left1 > right1:
        #    return right2 - left2 + 1
        #if left2 < 0 or left2 > right2:
        #    return right1 - left1 + 1
        #if memo[left1, right1, left2, right2] >= 0:
        #    return memo[left1, right1, left2, right2]
        if (left1, right1, left2, right2) in memo:
            return memo[left1, right1, left2, right2]
        # v1: modify
        if indexToLabel1[right1] == indexToLabel2[right2]:
            # right top nodes are same
            v1 = 0
        else:
            v1 = 1
        # distance of right most forest
        v1 += calcInnerDistance(
                indexToLeftRange1[right1], right1-1,
                indexToLeftRange2[right2], right2-1,
              )
        # distance of left neighbor forest
        v1 += calcInnerDistance(
                left1, indexToLeftRange1[right1]-1,
                left2, indexToLeftRange2[right2]-1,
              )
        # v2: remove
        v2 = calcInnerDistance(
                left1, right1,
                left2, right2-1
              ) + 1
        # v3: add
        v3 = calcInnerDistance(
                left1, right1-1,
                left2, right2
              ) + 1
        # v: minimal distance
        v = min(v1, v2, v3)
        memo[left1, right1, left2, right2] = v
        #pprint.pprint(memo)
        #print("%s,%s :: %s,%s" % (indexToLabel1[left1:right1+1], indexToLeftRange1[left1:right1+1], indexToLabel2[left2:right2+1], indexToLeftRange2[left2:right2+1]))
        #print("(%s,%s) vs (%s,%s) -> min(%s,%s,%s) = %s" % (left1, right1, left2, right2, v1, v2, v3, v))
        #print("(%s,%s) vs (%s,%s) -> %s" % (left1, right1, left2, right2, v))
        #print("------")
        return v
    #return calcInnerDistance(0, len(indexToLabel1)-1, 0, len(indexToLabel2)-1)
    return calcInnerDistance(0, numNodes1-1, 0, numNodes2-1)





#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''glue rules extracting function'''

# Standard libraries
import argparse
import sys
import subprocess

# Local libraries
from lpu.common import files
from lpu.common.progress import view
from lpu.smt.trans_models import records

def makeGlueRules(srcRuleTable, saveRules, progress = False):
    setTags = set()
    setPOSTags = set()
    if type(srcRuleTable) == str:
      #srcRuleTable = files.open(srcRuleTable)
      srcRuleTable = files.open(srcRuleTable, 'rt')
    if type(saveRules) == str:
        saveRules = open(saveRules, 'wt')
    if progress:
        srcRuleTable = view(srcRuleTable)
    for line in srcRuleTable:
        rec = records.TravatarRecord(line)
        symbols = rec.src.split(' ')
        if len(symbols) == 3:
            if symbols[0][0] == '"' and symbols[1] == "@":
                # single source word rule
                tag = symbols[2]
                setPOSTags.add(tag)
        elif len(symbols) >= 3 and symbols[-2] == "@":
            tag = symbols[-1]
            setTags.add(tag)

    setTags.add('X')
    for tag in setTags:
        if tag != "S":
            saveRules.write("x0:%s @ S ||| x0:%s @ S ||| glue=1\n" % (tag, tag))
    for tag in setPOSTags:
        if tag != "X":
            saveRules.write("x0:X @ %s ||| x0:X @ %s ||| glue=1 unk=1\n" % (tag, tag))
    #saveRules.write("x0:X x1:X @ X ||| x0:X x1:X @ X ||| glue=1\n")
    saveRules.write("x0:X x1:X @ X ||| x0:X x1:X @ X ||| glue=1 unk=1\n")

def main():
    parser = argparse.ArgumentParser(
        description='generate glue rules from travatar rule table'
    )
    parser.add_argument('srcRuleTable',  help='file path to load rule-table')
    parser.add_argument('saveRules', help='file path to save glue rules')
    parser.add_argument('--progress', '-p', action='store_true',
                        help='show progress bar (pv command should be installed')
    args = vars(parser.parse_args())
    #print(args)
    makeGlueRules(**args)

if __name__ == '__main__':
    main()




#!/usr/bin/env python

import argparse
import sys

def guessLangCodeFromFileName(filepath):
    fields = filepath.split('.')
    fields.reverse()
    for field in fields:
        if len(field) == 2:
            return field.lower()
    return "UNK"

def printLangCodeList(filepaths):
    print(str.join(' ', map(guessLangCodeFromFileName, filepaths)))

def cmdGuessLangCode(args):
    parser = argparse.ArgumentParser(description='Guess the language codes from given files')
    parser.add_argument('filepaths', metavar="filepath", nargs="+", type=str, help='path of file to guess the language code')
    #parser.add_argument('--from-filename', '-n', action='store_true', help='guess from the file name (default)')
    parsed = parser.parse_args(args)
    printLangCodeList(parsed.filepaths)

def main():
    cmdGuessLangCode(sys.argv[1:])

if __name__ == '__main__':
    main()




# distutils: language=c++
# -*- coding: utf-8 -*-

# Standard libraries
import argparse

# 3-rd party library
import numpy as np
cimport numpy as np

# Local libraries

from lpu.common import environ
from lpu.common import files
from lpu.common import progress
from lpu.common import logging

#from lpu.common.config import Config
from lpu.common.config cimport Config
from lpu.common.vocab cimport StringEnumerator

logger = logging.getLogger(__name__)

ITERATION_LIMIT = 10

NULL_SYMBOL = '__NULL__'

cdef class Vocab:
    # imported from "ibm_model1.pxd"
    #cdef StringEnumerator src
    #cdef StringEnumerator trg

    def __cinit__(self):
        self.init()
    cdef void init(self):
        self.src = StringEnumerator()
        self.trg = StringEnumerator()

    cpdef tuple ids_pair_to_str_pair(self, src_ids, trg_ids):
        cdef src_str = str.join(' ', [self.src.id2str(i) for i in src_ids[:-1]])
        cdef trg_str = str.join(' ', [self.trg.id2str(i) for i in trg_ids])
        return src_str, trg_str

    cpdef list load_sent_pairs(self, str src_path, str trg_path):
        cdef str src_line, trg_line
        cdef list src_words, trg_words
        cdef list src_ids, trg_ids
        cdef list sent_pairs
        cdef str word
        logger.info("loading files: %s %s" % (src_path,trg_path))
        self.src.append(NULL_SYMBOL)
        src_file = progress.view(files.open(src_path), 'loading')
        trg_file = files.open(trg_path)
        sent_pairs = []
        for src_line, trg_line in zip(src_file, trg_file):
            src_words = src_line.rstrip("\n").split(' ')
            trg_words = trg_line.rstrip("\n").split(' ')
            src_words.append(NULL_SYMBOL)
            src_ids = [self.src.str2id(word) for word in src_words]
            trg_ids = [self.trg.str2id(word) for word in trg_words]
            sent_pairs.append( (src_ids, trg_ids) )
        return sent_pairs

cdef tuple grid_indices(list x_indices, list y_indices):
    cdef np.ndarray indices1, indices2
    indices1, indices2 = np.meshgrid(x_indices, y_indices, sparse=True)
    return indices1.T, indices2.T

cdef np.ndarray sub_matrix(np.ndarray matrix, list x_indices, list y_indices):
    return matrix[grid_indices(x_indices, y_indices)]

#cdef void add_sub_matrix(np.ndarray target_matrix, list x_indices, list y_indices, np.ndarray source_matrix):
#    cdef np.ndarray indices1, indices2
#    indices1, indices2 = grid_indices(x_indices, y_indices)
#    np.add.at(target_matrix, (indices1,indices2), source_matrix[indices1,indices2])

cdef class Model:
    # imported from "ibm_model1.pxd"
    #cdef np.ndarray trans_dist
    #cdef Vocab vocab

    def __cinit__(self):
        self.init()
    cdef inline void init(self):
        self.vocab = Vocab()

    #cpdef double calc_pair_entropy(self, list src_sent, list trg_sent):
    cpdef double calc_pair_entropy(self, list src_sent, list trg_sent) except *:
        cdef np.ndarray trans_matrix
        trans_matrix = sub_matrix(self.trans_dist, src_sent, trg_sent)
        #return -np.log(trans_matrix.sum(axis=0) / len(src_sent)).sum()
        return (-np.log(trans_matrix.sum(axis=0) / len(src_sent)).sum()) / len(trg_sent)

    cpdef double calc_entropy(self, list sent_pairs) except *:
        cdef float total_entropy = 0
        logger.info('calculating entropy')
        for i, (src_sent, trg_sent) in enumerate(progress.view(sent_pairs, 'progress')):
            total_entropy += self.calc_pair_entropy(src_sent, trg_sent)
        return total_entropy / len(sent_pairs)

    cpdef void calc_and_save_scores(self, out_path, list sent_pairs):
        cdef double entropy = 0
        cdef str src_string, trg_string
        cdef list src_ids, trg_ids
        cdef str record
        logger.info("calculating and storing into file: %s"  % (out_path))
        with files.open(out_path, 'wt') as fobj:
            for i, (src_ids, trg_ids) in enumerate(progress.view(sent_pairs, 'progress')):
                entropy = self.calc_pair_entropy(src_ids, trg_ids)
                src_string, trg_string = self.vocab.ids_pair_to_str_pair(src_ids, trg_ids)
                record = "%s\t%s\t%s\n" % (entropy, src_string, trg_string)
                fobj.write(record)

    cpdef void save_align(self, out_path, threshold):
        cdef tuple indices
        cdef int src, trg
        cdef float prob
        cdef str record
        with files.open(out_path, 'wt') as fobj:
            logger.info("storing translation probabilities into file (threshold=%s): %s" % (threshold,out_path))
            indices = np.where(self.trans_dist>= threshold)
            for src, trg in progress.view(zip(*indices), 'storing', max_count=len(indices[0])):
                prob = self.trans_dist[src,trg]
                record = "%s\t%s\t%s\n" % (self.vocab.src.id2str(src), self.vocab.trg.id2str(trg), prob)
                fobj.write(record)

cdef class Trainer:
    # imported from "ibm_model1.pxd"
    #cdef Model model
    #cdef str src_path
    #cdef str trg_path
    #cdef list sent_pairs
    #cdef np.ndarray cooc_src_trg

    def __cinit__(self, conf, **others):
        self.init()
        conf = Config(conf)
        conf.update(**others)
        if conf.has(['src_path', 'trg_path']):
            self.src_path = conf.data.src_path
            self.trg_path = conf.data.trg_path
    cdef void init(self):
        self.model = Model()

    cpdef np.ndarray calc_uniform_dist(self):
        cdef StringEnumerator vocab_src = self.model.vocab.src
        cdef StringEnumerator vocab_trg = self.model.vocab.trg
        cdef np.ndarray uniform_dist
        cdef int src, trg
        logger.info("calculating uniform distribution for word translation probability")
        uniform_dist = np.ones([len(vocab_src), len(vocab_trg)], np.float64) / len(vocab_trg)
        logger.info("word trans. distribution matrix size: %s [src words] x %s [trg words] x %s [bytes] = %s [bytes]"
            % (uniform_dist.shape[0],uniform_dist.shape[1],uniform_dist.itemsize,len(uniform_dist.data)))
        return uniform_dist

    cpdef np.ndarray count_cooccurrence(self, sent_pairs):
        cdef StringEnumerator vocab_src = self.model.vocab.src
        cdef StringEnumerator vocab_trg = self.model.vocab.trg
        cdef np.ndarray cooc_src_trg
        cdef int src, trg
        cdef np.ndarray src_indices, trg_indices
        cooc_src_trg = np.zeros([len(vocab_src),len(vocab_trg)], np.int)
        logger.info('counting co-occurrences of source word and target word')
        for i, (src_sent, trg_sent) in enumerate(progress.view(sent_pairs, 'processing')):
            #np.add.at(cooc_src_trg, np.meshgrid(src_sent,trg_sent), 1)
            np.add.at(cooc_src_trg, grid_indices(src_sent, trg_sent), 1)
        logger.info("co-occurrence matrix size: %s [src words] x %s [trg words] x %s [bytes]= %s [bytes]"
            % (cooc_src_trg.shape[0],cooc_src_trg.shape[1],cooc_src_trg.itemsize,cooc_src_trg.size*cooc_src_trg.itemsize))
        return cooc_src_trg

    cpdef void train(self, int iteration_limit):
    #cpdef void train(self, int iteration_limit) except *:
        cdef double last_entropy
        logger.info("start training IBM Model 1")
        self.train_first()
        last_entropy = self.model.calc_entropy(self.sent_pairs)
        logger.info("initial entropy: %s" % last_entropy)
        for step in range(iteration_limit):
            logger.info("--")
            logger.info("step: %s" % (step + 1))
            # train 1 step
            self.train_step()
            # calculate entropy
            entropy = self.model.calc_entropy(self.sent_pairs)
            logger.info("entropy: %s" % entropy)
            if entropy == last_entropy:
                break
            else:
                last_entropy = entropy

    cpdef void train_first(self):
        self.sent_pairs = self.model.vocab.load_sent_pairs(self.src_path, self.trg_path)
        logger.info("source vocabulary size: {:,d}".format(len(self.model.vocab.src)))
        logger.info("target vocabulary size: {:,d}".format(len(self.model.vocab.trg)))
        self.cooc_src_trg = self.count_cooccurrence(self.sent_pairs)
        self.model.trans_dist = self.calc_uniform_dist()

    cpdef void train_step(self):
        cdef long i
        cdef list src_sent, trg_sent
        cdef np.ndarray count_src2trg, total_src
        cdef np.ndarray trg_factor
        cdef StringEnumerator vocab_src = self.model.vocab.src
        cdef StringEnumerator vocab_trg = self.model.vocab.trg
        cdef tuple grid

        if len(self.model.trans_dist) == 0:
            self.train_first()
        else:
            count_src2trg = np.zeros([len(vocab_src),len(vocab_trg),], np.float64)
            total_src = np.zeros(len(vocab_src), np.float64)
            logger.info("estimating expected co-occurrence counts")
            for i, (src_sent, trg_sent) in enumerate(progress.view(self.sent_pairs, 'processing')):
                grid = grid_indices(src_sent, trg_sent)
                # compute normalization
                trg_factor = np.zeros(len(vocab_trg), np.float64)
                np.add.at(trg_factor, trg_sent, sub_matrix(self.model.trans_dist,src_sent,trg_sent).sum(axis=0))
                # collect counts
                np.add.at(count_src2trg, grid,     self.model.trans_dist[grid] / trg_factor[trg_sent])
                np.add.at(total_src, src_sent, (self.model.trans_dist[grid] / trg_factor[trg_sent]).sum(axis=1))
            # estimate probabilities
            logger.info("estimating word translation probabilities")
            self.model.trans_dist = count_src2trg / total_src.reshape([-1,1])

def check_config(conf):
    logger.debug(conf)

def train_ibm_model1(conf, **others):
    with environ.push() as e:
        conf = Config(conf)
        conf.update(others)
        check_config(conf)
        trainer = Trainer(conf, **others)
        try:
            #np.seterr(all='raise')
            #trainer.train(conf.data.iteration_limit)
            with np.errstate(all='raise'):
                trainer.train(conf.data.iteration_limit)
        except KeyboardInterrupt as k:
            logger.debug(k)
            logger.info("forcing to dump alignments and scores")
        #except Exception as e:
        #    logger.debug(e)
        #    logger.info("forcing to dump alignments and scores")
        trainer.model.save_align(conf.data.save_align_path, conf.data.threshold)
        if conf.data.save_scores:
            trainer.model.calc_and_save_scores(conf.data.save_scores, trainer.sent_pairs)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('src_path', metavar='src_path (in)', help='file containing source-side lines of parallel text', type=str)
    parser.add_argument('trg_path', help='file containing target-side lines of parallel text', type=str)
    parser.add_argument('save_align_path', help='output file to save alignment', type=str)
    parser.add_argument('--save-scores', '-S', help='output file to save entropy of each each alignment', type=str, default=None)
    parser.add_argument('--iteration-limit', '-I', help='maximum iteration number of EM algorithm (default: %(default)s)', type=int, default=ITERATION_LIMIT)
    parser.add_argument('--threshold', '-t', help='threshold of translation probabilities to save', type=float, default=0.01)
    #parser.add_argument('--character', '-c', help='chacacter based alignment mode', action='store_true')
    parser.add_argument('--debug', '-D', help='debug mode', action='store_true')
    parser.add_argument('--quiet', '-q', help='not showing staging log', action='store_true')
    args = parser.parse_args()
    conf = Config(vars(args))
    with logging.push_environ(logger) as e:
        if conf.data.debug:
            e.set_debug(True)
            e.set_quiet(False)
        if conf.data.quiet:
            e.set_quiet(True)
            e.set_debug(False)
        logger.debug(conf)
        train_ibm_model1(conf)

if __name__ == '__main__':
    main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Standard libraries
import argparse
import os
import re
import sys
import unicodedata
from functools import reduce

# Local libraries
from lpu.common import compat
from lpu.common import logging
from lpu.common import progress

logger = logging.getColorLogger(__name__)

REPLACE_MAP = {
    compat.to_unicode('<'): compat.to_unicode('-LT-'),
    compat.to_unicode('>'): compat.to_unicode('-GT-'),
    compat.to_unicode('('): compat.to_unicode('-LRB-'),
    compat.to_unicode(')'): compat.to_unicode('-RRB-'),
    compat.to_unicode('{'): compat.to_unicode('-LCB-'),
    compat.to_unicode('}'): compat.to_unicode('-RCB-'),
    compat.to_unicode('['): compat.to_unicode('-LSB-'),
    compat.to_unicode(']'): compat.to_unicode('-RSB-'),
    compat.to_unicode('|'): compat.to_unicode('-BAR-'),
    compat.to_unicode('&'): compat.to_unicode('-AMP-'),
    compat.to_unicode('\t'): compat.to_unicode(' '),
    unicodedata.lookup('ZERO WIDTH SPACE'): compat.to_unicode(' '),
    unicodedata.lookup('ZERO WIDTH NO-BREAK SPACE'): compat.to_unicode(' '),
}

def getLongestCommonPrefix(s1, s2):
    index = 0
    f1 = s1.split('.')
    f2 = s2.split('.')
    while True:
        if len(f1) <= index or len(f2) <= index:
            break
        if f1[index] != f2[index]:
            break
        index += 1
    return str.join('.', f1[0:index])

def getLongestCommonSuffix(s1, s2):
    return getLongestCommonPrefix(s1[::-1],s2[::-1])[::-1]

def replaceChar(c):
    if c in REPLACE_MAP:
        #logging.log("Replacing '%s' -> '%s'" % (c, REPLACE_MAP[c]))
        return REPLACE_MAP[c]
    else:
        return c

def normalize(line, escape=False):
    #line = compat.to_unicode( line.strip() )
    line = unicodedata.normalize('NFKD', line)
    if escape:
        line = compat.to_unicode('').join(map(replaceChar, line))
    line = unicodedata.normalize('NFC', line)
    line = compat.to_str(line)
    line = re.sub(r'\s+', ' ', line)
    return line

def checkLength(lines, minLength, maxLength):
    for line in lines:
        words = line.split()
        if len(words) < minLength: return False
        if len(words) > maxLength: return False
    return True

def getDiff(s, prefix, suffix):
    if len(suffix) == 0:
        return s[len(prefix):None]
    else:
        return s[len(prefix):-len(suffix)]

def cleanParallel(**args):
    srcFilePaths = args.get('srcFilePaths')
    #outPrefix = args.get('outfileprefix')
    outTag = args.get('outTag')
    minLength = args.get('min')
    maxLength = args.get('max')
    out_dir    = args.get('target_directory')

    if not os.path.isdir(out_dir):
        logging.log("Making directory: %s" % out_dir)
        os.makedirs(out_dir)

    #print(args)
    srcBaseNames = list( map(os.path.basename, srcFilePaths) )
    commonPrefix = reduce(getLongestCommonPrefix, srcBaseNames)
    commonSuffix = reduce(getLongestCommonSuffix, srcBaseNames)
    logger.info("common prefix: '{}'".format(commonPrefix))
    logger.info("common suffix: '{}'".format(commonSuffix))
    outPaths = []
    for i, path in enumerate(srcFilePaths):
        if outTag[0:1] != '.':
            outTag = '.' + outTag
        if commonSuffix:
            outPath = path + outTag
        else:
            diff = getDiff(os.path.basename(path), commonPrefix, commonSuffix)
            logger.info("suffix {}: {}".format(i+1, diff))
            outPath = commonPrefix + outTag + diff
        outPaths.append(os.path.join(out_dir, outPath))
    logger.info("writing cleaned corpora into: %s ..." % str.join(' ',outPaths))
    if sys.version_info.major >= 3:
        infiles  = [open(path,'rb') for path in srcFilePaths]
    else:
        infiles  = [open(path,'r') for path in srcFilePaths]
    outfiles = [open(path,'w') for path in outPaths]
    infiles[0] = progress.view(infiles[0], header='processing')
    for i, lines in enumerate(zip(*infiles)):
        try:
            lines = [compat.to_unicode(line.strip()) for line in lines]
            if args.get('normalize'):
                escape = args.get('escape')
                #lines = list( map(normalize, lines) )
                lines = [normalize(line, escape=escape) for line in lines]
            if checkLength(lines, minLength, maxLength):
                for i, line in enumerate(lines):
                    outfiles[i].write(line)
                    outfiles[i].write("\n")
        except Exception as e:
            #sys.stdout.write("\n")
            logging.warn("%s (Line %s)" % (e, i))

def main():
    DEFAULT_MIN_LENGTH = 1
    DEFAULT_MAX_LENGTH = 80
    DEFAULT_RATIO = 9.0
    parser = argparse.ArgumentParser(description='Clean parallel corpus by length and normalize Unicode chars')
    parser.add_argument('srcFilePaths', metavar="filepath", nargs="+", type=str, help='path of file to clean')
    parser.add_argument('outTag', metavar="output_tag", type=str, help='tag added in name of file to save')
    parser.add_argument('--min', default=DEFAULT_MIN_LENGTH, type=int, help='minimum #words per line (default: %(default)s)')
    parser.add_argument('--max', default=DEFAULT_MAX_LENGTH, type=int, help='maximum #words per line (default: %(default)s)')
    parser.add_argument('--ratio', default=DEFAULT_RATIO, type=float, help='upper bound of maximal ratio of #words between each 2 lines (default: %(default)s)')
    parser.add_argument('--target-directory', '-D', default='./', type=str, help='directory to save the cleaned texts (default: %(default)s)')
    parser.add_argument('--escape', action='store_true', help='escape special characters (default: %(default)s)')
    parser.add_argument('--normalize', action='store_true', help='perform unicode normalization (default: %(default)s)')
    #parsed = parser.parse_args(args)
    parsed = parser.parse_args()
    logger.debug(parsed)
    cleanParallel(**vars(parsed))

if __name__ == '__main__':
    main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''Auxiliary functions for file I/O'''

# Standard libraries
import gzip
import io
import os.path
import sys
import tempfile

# Local libraries
from lpu.common import logging

logger = logging.getLogger(__name__)

#env = os.environ
#env['LC_ALL'] = 'C'

DEFAULT_BUFFER_SIZE = 10 * (1024 ** 2) # 10MB

_open = open

if sys.version_info.major >= 3:
    bin_stdin  = sys.stdin.buffer
    bin_stdout = sys.stdout.buffer
    bin_stderr = sys.stderr.buffer
    FileType = io.IOBase
else:
    bin_stdin  = sys.stdin
    bin_stdout = sys.stdout
    bin_stderr = sys.stderr
    FileType = file
    FileNotFoundError = IOError

#def autoCat(filenames, target):
def concat_into(filenames, target, progress=True):
    '''
        concatenate and copy files into target file (expanding for compressed ones)
    '''
    if type(filenames) != list:
        filenames = [filenames]
    f_out = open(target, 'wb')
    if progress:
        from lpu.common.progress import view as pview
        logger.info("writing into: {}".format(target))
        for filename in filenames:
            f_in = pview(filename, header="  transfering from: {}".format(filename))
            for line in f_in.read_byte_chunks(DEFAULT_BUFFER_SIZE):
                f_out.write(line)
            f_in.close()
    else:
        for filename in filenames:
            f_in = open(filename, 'rb')
            for line in f_in:
                f_out.write(line)
            f_in.close()
    f_out.close()

def castFile(anyFile):
    '''try to convert any argument to file like object
    * if given file like object, then return itself
    * otherwise (e.g. given file path string), try to open it and return file object'''
    if hasattr(anyFile, 'read'):
        return anyFile
    else:
        return open(anyFile)

def getContentSize(path):
    '''get the file content size (expanded size for compressed one)'''
    try:
        f_in = _open(path, 'rb')
        if is_gzipped(path):
            f_in.seek(-8, 2)
            crc32 = gzip.read32(f_in)
            isize = gzip.read32(f_in)
            f_in.close()
            return isize
        else:
            f_in.seek(0, 2)
            pos = f_in.tell()
            f_in.close()
            return pos
    except Exception as e:
        return -1


def get_ext(filename):
    '''get the extension of given file'''
    (name, ext) = os.path.splitext(filename)
    return ext

#def isIOType(obj):
#    #return isinstance(obj, (io.IOBase,file))
#    return isinstance(obj, FileType)

def is_gzipped(filename):
    '''check whether the given file is compressed by gzip or not'''
    try:
        f = gzip.open(filename, 'r')
        f.readline()
        return True
    except Exception as e:
        return False

def is_mode(fobj, mode):
    if mode in ('r', 'read'):
        return fobj.mode.find('r') >= 0
    elif mode in ('w', 'write'):
        return fobj.mode.find('w') >= 0
    elif mode in ('b', 'binary'):
        if fobj.mode.find('b') >= 0:
            return True
        elif fobj.mode.find('t') >= 0:
            return False
        else:
            return sys.version_info.major < 3
    elif mode in ('t', 'text'):
        if fobj.mode.find('t') >= 0:
            return True
        elif fobj.mode.find('b') >= 0:
            return False
        else:
            return sys.version_info.major >= 3

def load(filepath_or_buffer, out_buffer, progress = True, bs = DEFAULT_BUFFER_SIZE):
    '''load all the content of given file (expand if compressed)'''
    if isinstance(filepath_or_buffer, str):
        # file path
        filepath = filepath_or_buffer
        buf = _open(filepath, 'rb')
        if progress:
            from lpu.common.progress import SpeedCounter
            header = "loading file '{}'".format(filepath)
            max_count = os.path.getsize(filepath)
            c = SpeedCounter(max_count=max_count, header=header)
    else:
        filepath = None
        buf = filepath_or_buffer
        if progress:
            from lpu.common.progress import SpeedCounter
            header = "loading buffer"
            c = SpeedCounter(header=header)
    #data = io.BytesIO()
    #f_in = _open(filename, 'rb')
    #if progress:
    #    from lpu.common.progress import SpeedCounter
    #    header = "loading file '%(filename)s'" % locals()
    #    max_count = os.path.getsize(filename)
    #    c = SpeedCounter(max_count=max_count, header=header)
    while True:
        #buf = f_in.read(bs)
        data = buf.read(bs)
        if not data:
            break
        else:
            #data.write(buf)
            out_buffer.write(data)
            if progress:
                #c.set_count(data.tell(), True)
                c.set_count(out_buffer.tell(), True)
    #f_in.close()
    buf.close()
    #data.seek(0)
    out_buffer.seek(0)
    if progress:
        c.reset()
    if isinstance(filepath, str) and get_ext(filepath) == '.gz':
        #f_in = gzip.GzipFile(fileobj = data)
        #f_in.myfileobj = data
        fobj = gzip.GzipFile(fileobj = out_buffer)
        fobj.myfileobj = out_buffer
        return fobj
    else:
        return out_buffer

def load_to_buffer(filepath_or_buffer, progress=True, bs=DEFAULT_BUFFER_SIZE):
    #data = io.BytesIO()
    return load(filepath_or_buffer, io.BytesIO(), progress, bs)

def load_to_temp(filepath_or_buffer, progress=True, bs=DEFAULT_BUFFER_SIZE, named = True):
    if named:
        temp = tempfile.NamedTemporaryFile()
        load(filepath_or_buffer, temp.file, progress, bs)
        return temp
    else:
        temp = tempfile.TemporaryFile()
        load(filepath_or_buffer, temp, progress, bs)
        return temp

def safeMakeDirs(dirpath, **options):
    '''make directories recursively for given path, don't throw exception if directory exists but if file exists'''
    if not os.path.isdir(dirpath):
        logger.debug('making directory: "%s"' % dirpath)
        try:
            os.makedirs(dirpath, **options)
        except:
            logger.debug('cannot make directory: "%s"' % dirpath)

def open(filename, mode = 'r'):
    '''open the plain/compressed file transparently'''
    if get_ext(filename) == '.gz':
        file_obj = gzip.open(filename, mode)
    elif mode.find('r') >= 0 and is_gzipped(filename):
        file_obj = gzip.open(filename, mode)
    else:
        #logger.debug("normal open mode")
        if mode.find('t') >= 0 and sys.version_info[0] >= 3:
            #file_obj = _open(filename, mode, encoding='utf-8', errors='replace')
            file_obj = _open(filename, mode, encoding='utf-8', errors='backslashreplace')
        else:
            file_obj = _open(filename, mode)
    return file_obj

def rawfile(f):
    if hasattr(f, 'myfileobj'):
        # for archive files such as gzip
        return f.myfileobj
#    if isinstance(f, gzip.GzipFile):
#        return f.myfileobj
    elif hasattr(f, 'buffer'):
        # for buffered files such as utf-8 mode
        #return f.buffer
        # might be duplicated buffer
        return rawfile(f.buffer)
    #elif isIOType(f):
    elif isinstance(f, FileType):
        return f
    else:
        logger.debug(f)
        #logging.debug(type(f))
        #logging.debug(dir(f))
        assert False

def rawsize(f):
    try:
        raw = rawfile(f)
        pos = raw.tell()
        raw.seek(-1, 2)
        size = raw.tell()
        raw.seek(pos, 0)
        return size
    except Exception as e:
        logger.debug(repr(e))
        return -1

def rawremain(f):
    return rawsize(f) - rawtell(f)

def rawtell(fileobj):
    '''get the current position of the opend file, return raw (not expanded) position for compressed'''
    return rawfile(fileobj).tell()

def testFile(path):
    '''test file existence'''
    if os.path.isfile(path):
        return True
    #logger.debug("file does not exist: '%s'" % path)
    raise FileNotFoundError("file does not exist: '%s'" % path)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

__all__ = [
    'common',
    'commands',
    'data_structs',
    'smt',
    '__system__',
]

# initializing

import os.path
from . common import logging

version_file = os.path.join(os.path.dirname(__file__), 'VERSION')
__version__ = open(version_file).read().strip()

logger = logging.getColorLogger(__name__)

if logging.get_quiet_status():
    logger.setLevel(logging.ERROR)
elif logging.get_debug_status():
    logger.setLevel(logging.DEBUG)
else:
    # verbose mode
    logger.setLevel(logging.INFO)

logger.debug("Initialized LPU")




# distutils: language=c++
# -*- coding: utf-8 -*-

## C++ setting
#from libcpp cimport bool
from libcpp.string cimport string
#from libcpp.vector cimport vector
from libcpp.deque cimport deque
#from libcpp.list cimport list as stl_list
#ctypedef unsigned char byte

import sys

import logging
logger = logging.getLogger(__name__)

# 3rd party library
try:
    import pycedar
except Exception as e:
    logger.exception(e)
    logger.error("module 'pycedar' is not found, please install 'pycedar' package (e.g. $ pip install --user pycedar)")
    sys.exit(1)

# Local libraries
from lpu.common import compat

'''Dictionaries and ID Maps implemented by Double-Array Trie'''

cdef class IDMap:
    '''auto mapping class from string to unique int'''
    cdef object dict
    cdef list unusedIDs
    cdef int numEmpty

    def __cinit__(self):
        #self.trie      = pycedar.trie()
        self.dict      = pycedar.dict()
        self.unusedIDs = list()
        self.numEmpty   = 0

    cpdef long append(self, str key):
        cdef long n = self.str2id(key)
        if not key:
            self.numEmpty = 1
            return 0
        elif n >= 0:
            return n
        else:
            if len(self.unusedIDs) > 0:
                n = self.unusedIDs.pop()
            else:
                n = self.dict.num_keys() + 1
            self.dict[key] = n
            return n

    def ids(self):
        if self.numEmpty > 0:
            yield 0
        for r in self.dict.predict(''):
            yield r.value()

    def items(self):
        if self.numEmpty > 0:
            yield ('', 0)
        for r in self.dict.predict(''):
            yield (r.key(), r.value())

    def keys(self):
        if self.numEmpty > 0:
            yield ''
        for r in self.dict.predict(''):
            yield r.key()

    cpdef long remove(self, str key):
        cdef long n = self.str2id(key)
        if n == 0:
            self.numEmpty = 0
            return 0
        elif n > 0:
            self.dict.erase(key)
            self.unusedIDs.append(n)
            return n
        else:
            raise KeyError(key)

    cpdef long str2id(self, str key):
        if not key:
            if self.numEmpty > 0:
                return 0
            else:
                return -1
        try:
            return self.dict[key]
        except:
            return -1

    def __delitem__(self, str key):
        self.remove(key)
    def __getitem__(self, str key):
        return self.append(key)
    def __has__(self, str key):
        return self.str2id(key) >= 0
    def __iter__(self):
        return self.keys()
    def __len__(self):
        return self.dict.num_keys() + self.numEmpty

cdef class TwoWayIDMap(IDMap):
    '''auto mapping class from string to unique int and vice versa'''
    #cdef list keyList
    cdef deque[string] keyList

    def __cinit__(self):
        IDMap.__init__(self)
        #self.keyList = [None]
        self.keyList.clear()
        self.keyList.push_back(b'')

    cpdef long append(self, str key):
        cdef size_t n = IDMap.append(self, key)
        if n >= self.keyList.size():
            self.keyList.push_back(compat.to_bytes(key))
        else:
            #self.keyList[n] = key
            self.keyList[n] = compat.to_bytes(key)
        return n

    cpdef str id2str(self, long num):
        #cdef object key = self.keyList[num]
        cdef string key
        if num == 0 and self.numEmpty > 0:
            return ''
        key = self.keyList[num]
        #if key:
        if not key.empty():
            #return key
            return compat.to_str(key)
        else:
            #raise IndexError(key)
            raise IndexError(compat.to_str(key))

    def ids(self):
        #cdef long i
        cdef size_t i
        cdef string k
        #cdef object k
        #for i, k in enumerate(self.keyList):
        #    if k is not None:
        #        yield i
        if self.numEmpty == 1:
            yield 0
        for i in range(1, self.keyList.size()):
            k = self.keyList[i]
            if not k.empty():
                yield k

    def items(self):
        #cdef long i
        cdef size_t i
        cdef object k
        #for i, k in enumerate(self.keyList):
        #    if k is not None:
        #        yield (k, i)
        for i in range(1, self.keyList.size()):
            k = self.keyList[i]
            if not k.empty():
                #yield (i, k)
                yield (i, compat.to_str(k))

    def keys(self):
        #cdef long i
        cdef size_t i
        cdef object k
        #for k in self.keyList:
        #    if k is not None:
        #        yield k
        for i in range(1, self.keyList.size()):
            k = self.keyList[i]
            if not k.empty():
                #yield k
                yield compat.to_str(k)

    cpdef void purge(self):
        while True:
            #if len(self.keyList) <= 1:
            if self.keyList.size() <= 1:
                break
            #if self.keyList[-1] is None:
            if self.keyList.back().empty():
                #self.keyList.pop()
                self.keyList.pop_back()
            else:
                break

    cpdef long remove(self, str key):
        cdef long n = IDMap.remove(self, key)
        if n >= 0:
            #self.keyList[n] = None
            self.keyList[n] = b''
        else:
            raise KeyError(key)

    def __delitem__(self, str key):
        self.remove(key)
    def __getitem__(self, str key):
        return self.append(key)
    def __iter__(self):
        return self.keys()

cdef class Dict:
    '''mapping class from string to any object'''
    cdef IDMap idmap
    cdef list objectList

    def __cinit__(self):
        self.idmap = IDMap()
        self.objectList = [None]

    cpdef object get(self, str key, object default=None):
        cdef long n = self.idmap.str2id(key)
        if n >= 0:
            return self.objectList[n]
        else:
            return default

    cpdef object remove(self, str key):
        cdef long n
        cdef object value
        n = self.idmap.remove(key)
        if n >= 0:
            value = self.objectList[n]
            self.objectList[n] = None
            return value
        else:
            raise KeyError(key)

    cpdef object setdefault(self, str key, object value):
        if key in self:
            return self[key]
        else:
            return self.__set(key, value)
    cdef object __set(self, str key, object value):
        cdef long n = self.idmap.append(key)
        if n >= len(self.objectList):
            self.objectList.append(value)
        else:
            self.objectList[n] = value
        return value

    def items(self):
        for key, i in self.idmap.items():
            yield (key, self.objectList[i])

    def keys(self):
        return self.idmap.keys()

    def values(self):
        for i in self.idmap.ids():
            yield self.objectList[i]

    def __delitem__(self, str key):
        self.remove(key)
    def __getitem__(self, str key):
        if key in self:
            return self.get(key)
        else:
            raise KeyError(key)
    def __setitem__(self, str key, object value):
        self.__set(key, value)

    def __has__(self, str key):
        if not key:
            raise KeyError(key)
        return key in self.idmap
    def __iter__(self):
        return self.keys()
    def __len__(self):
        return len(self.idmap)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Standard libraries
import argparse
import random

# Local libraries
from lpu.common import compat
from lpu.common import files
from lpu.common import environ
from lpu.common import logging
from lpu.common import numbers
from lpu.common import progress
from lpu.common.config import Config

logger = logging.getColorLogger(__name__)

def get_valid_indices(conf):
    #indices = set()
    indices = []
    infiles = [files.open(path,'rb') for path in conf.data.inpaths]
    infiles[0] = progress.view(infiles[0], 'loading')
    #for i, lines in enumerate(progress.view(compat.zip(*infiles), 'loading')):
    for i, lines in enumerate(compat.zip(*infiles)):
        try:
            lines = map(compat.to_unicode, lines)
            if conf.data.ignore_empty:
                if all([line.rstrip("\n") for line in lines]):
                        indices.append(i)
            else:
                indices.append(i)
        except Exception as e:
            #sys.stdout.write("\n")
            logger.warning("%s (Line %s)" % (e, i))
    return indices

def random_split(conf, **others):
    conf = Config(conf, **others)
    if not check_config(conf):
        return False

    inpaths = conf.data.inpaths
    prefixes = conf.data.prefixes
    suffixes = conf.data.suffixes
    tags = conf.data.tags
    split_sizes = conf.data.split_sizes

    for path in inpaths:
        files.testFile(path)
    logger.info('building sequence')
    indices = get_valid_indices(conf)
    logger.info('loaded: {} lines'.format(len(indices)))
    if conf.data.seed >= 0:
        random.seed(conf.data.seed)
    logger.info("randomizing sequence")
    random.shuffle(indices)
    split_indices = []
    total_size = 0
    total_factor = 0
    for i, size in enumerate(split_sizes):
        if size == '*':
            total_factor += 1
        else:
            size = float(size)
            if size < 1:
                size = int(len(indices) * size)
            split_sizes[i] = size
            total_size += size
    start = 0
    for size in split_sizes:
        if size == '*':
            size = int( float(len(indices) - total_size) / total_factor )
        else:
            size = int(size)
        split_indices.append( set(indices[start:start+size]) )
        start += size
    split_sizes = list( map(len, split_indices) )
    for inpath, prefix, suffix in zip(inpaths, prefixes, suffixes):
        infile = files.open(inpath, 'rb')
        outpaths = []
        for tag in tags:
            if suffix:
                outpath = "%s%s.%s" % (prefix, tag, suffix)
            else:
                outpath = "%s%s" % (prefix, tag)
            outpaths.append(outpath)
        logger.info("writing lines into splitted files: %s" % outpaths)
        logger.info("split sizes: %s" % split_sizes)
        outfiles = [files.open(outpath,'wb') for outpath in outpaths]
        for line_index, line in enumerate(progress.view(infile, 'processing')):
            for file_index, outfile in enumerate(outfiles):
                if line_index in split_indices[file_index]:
                    outfile.write(line)
    if conf.data.ids:
        for file_index, tag in enumerate(tags):
            outpath = "%s.%s" % (tag, conf.data.ids)
            genIDs = iter(sorted(split_indices[file_index]))
            if progress:
                genIDs = progress.view(genIDs, "Writing IDs into '%s'" % outpath)
            outfile = files.open(outpath, 'wt')
            for line_index in genIDs:
                # line index is zero origin, line id should be +1
                outfile.write("%d\n" % (line_index+1))
    return True

def check_config(conf):
    numInput = len(conf.data.input)
    numTags  = len(conf.data.tags)
    conf.data.inpaths = conf.data.input
    conf.data.seed        = conf.data.random_seed
    conf.data.split_sizes = conf.data.split_sizes
    if not conf.data.prefixes and not conf.data.suffixes:
        logger.warning('at least one of prefixes or suffixes should be given')
        return False
    if not conf.data.prefixes:
        conf.data.prefixes = [''] * numInput
    elif len(conf.data.prefixes) != numInput:
        msg="Number of prefixes should be same with input files (expected %d, but given %d)"
        logger.warning(msg % (numInput, len(conf.data.prefixes)))
        return False
    if not conf.data.suffixes:
        conf.data.suffixes = [''] * numInput
    elif len(conf.data.suffixes) != numInput:
        msg="Number of suffixes should be same with input files (expected %d, but given %d)"
        logger.warning(msg % (numInput, len(conf.data.suffixes)))
        return False
    if len(conf.data.split_sizes) != numTags:
        msg="Number of split sizes should be same with input files (expected %d, but given %d)"
        logger.warning(msg % (numTags, len(conf.data.split_sizes)))
        return False
    for i, size in enumerate(conf.data.split_sizes):
        if size == '*':
            pass
        else:
            try:
                n = float(size)
                conf.data.split_sizes[i] = n
            except Exception as e:
                msg = "string '%s' cannnot be converted to number (given invalid split sizes: %s)"
                logger.warning(msg % (size, conf.data.split_sizes))
            if n <= 0:
                msg = "split size should be positive, but given negative: %s (given invalid split sizes: %s)"
                logger.warning(msg % (n, conf.data.split_sizes))
    logger.debug(conf)
    return True

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--input', '-I', help='input files', type=str, required=True, nargs='+')
    parser.add_argument('--prefixes', '-P', help='prefixes of splitted files (should be same number with INPUT)', type=str, default=[], nargs='+')
    parser.add_argument('--suffixes', '-S', help='suffixes of splittted files (should be same number with INPUT)', type=str, default=[], nargs='+')
    parser.add_argument('--tags', '-T', help='base names of splitted files (comma separated list)', type=str, required=True, nargs='+')
    parser.add_argument('--split-sizes', '-s', help='number of lines in splitted files (should be same number with TAGS)', type=str, required=True, nargs='+')
    parser.add_argument('--ignore-empty', '-E', help='preventing empty lines to output', action='store_true')
    parser.add_argument('--quiet', '-q', help='not showing staging log', action='store_true')
    parser.add_argument('--debug', '-D', help='verbose mode (including debug info)', action='store_true')
    parser.add_argument('--random-seed', '-R', help='random seed', default=-1, type=int)
    parser.add_argument('--ids', metavar='SUFFIX', help='write original line numbers for each tags', default=0, nargs='?', const='ids')
    args = parser.parse_args()
    conf = Config()
    conf.update(vars(args))
    with logging.using_config(logger) as c:
        if conf.data.debug:
            c.set_debug(True)
            c.set_quiet(False)
        if conf.data.quiet:
            c.set_quiet(True)
            c.set_debug(False)
        logger.debug(conf)
        random_split(conf)

if __name__ == '__main__':
    main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''functions to triangulate 2 phrase tables into 1 table
by combining source-pivot and pivot-target for common pivot phrase'''

# Standarde libraries
import argparse
import math
import sys

# Local libraries
from lpu.common import compat
from lpu.common import files
from lpu.common import logging
from lpu.common import progress
from lpu.smt.trans_models.records import MosesRecord, TravatarRecord
from lpu.smt.trans_models.tables import Table

from lpu.data_structs import trees

# lower threshold of trans probs to abort
#THRESHOLD = 1e-3
THRESHOLD = 0 # not aborting

# lower guarantee for lexical trans probs
MINPROB = 10 ** -10

# limit number of records for the same source phrase
NBEST = 20

# methods to estimate trans probs (countmin/prodprob/bidirmin/bidirgmean/bidirmax/bidiravr)
method_list = ['countmin', 'prodprob', 'bidirmin', 'bidirgmean', 'bidirmax', 'bidiravr']
#METHOD = 'counts'
#METHOD = 'hybrid'
#METHOD = 'countmin'
METHOD = 'prodprob'

# methods to estimate lexical weight
lex_method_list = ['prodweight', 'countmin', 'prodprob', 'bidirmin', 'bidirgmean', 'bidirmax', 'bidiravr', 'table', 'countmin+table', 'prodprob+table', 'bidirmin+table', 'bidirgmean+table']
LEX_METHOD = 'prodweight'

# methods to estimate joint trans probs
joint_method_list = ['memoryless', 'independent']
JOINT_METHOD = 'memoryless'

# matching methods
match_method_list = ['hiero', 'symbols', 'treecomp', 'treedist', 'treedistexp']
MATCH_METHOD = 'symbols'

NULLS = 10**4

NOPREFILTER = False

class WorkSet:
    '''data set for multi-processing'''
    def __init__(self, savefile, workdir, method, **options):
        prefix = options.get('prefix', 'phrase')
        self.multi_target = options.get('multi_target', False)
        self.Record = options.get('RecordClass', MosesRecord)
        self.method = method
        self.matchMethod = options.get('matchMethod', MATCH_METHOD)
        self.jointMethod = options.get('jointMethod', JOINT_METHOD)
#        if method.find('multi') >= 0:
#            self.multi_target = True
#            self.method = method.replace('multi','').replace('+','')
        self.nbest = NBEST
        self.savePath = savefile
        self.threshold = THRESHOLD
        self.workdir = workdir
        self.foutPivot = files.open(savefile, 'wt')
        #self.pivotProc = multiprocessing.Process( target = pivotRecPairs, args = (self,) )
        #self.recordProc = multiprocessing.Process( target = writeRecordQueue, args = (self,) )
        self.numRecSrcPvt = 0
        self.setPhrasesSrcPvt = set()
        self.setWordsSrcPvt = set()
        self.numRecPvtTrg = 0
        self.setPhrasesPvtTrg = set()
        self.setWordsPvtTrg = set()
        self.numRecSrcTrg = 0
        self.setPhrasesSrcTrg = set()
        self.setWordsSrcTrg = set()

    def __del__(self):
        self.close()

    def close(self):
        if self.foutPivot:
            self.foutPivot.close()
            self.foutPivot = None
#        if self.pivotProc.pid:
#            if self.pivotProc.exitcode == None:
#                self.pivotProc.terminate()
#            self.pivotProc.join()
#        if self.recordProc.pid:
#            if self.recordProc.exitcode == None:
#                self.recordProc.terminate()
#            self.recordProc.join()
#        self.pivotQueue.close()
#        self.outQueue.close()

#    def join(self):
#        self.pivotProc.join()
#        self.recordProc.join()
#
#    def start(self):
#        self.pivotProc.start()
#        self.recordProc.start()

#    def terminate(self):
#        self.pivotProc.terminate()
#        self.recordProc.terminate()


#def updateFeatures(recPivot, recPair, method, multi_target = False, jointMethod = 'memoryless'):
def updateFeatures(recPivot, recPair, workset, multi_target = False):
    '''update features'''
    features = recPivot.features
    srcFeatures = recPair[0].features
    trgFeatures = recPair[1].features
    #logging.debug(method)
    #logging.debug(srcFeatures)
    #logging.debug(trgFeatures)
    #if method.find('prodprob') >= 0:
    if workset.method.find('prodprob') >= 0:
        # multiplying scores and marginalizing
        if not multi_target:
            for key in ['egfl', 'egfp', 'fgel', 'fgep']:
                features.setdefault(key, 0)
                if workset.matchMethod == 'hiero':
                    features[key] += (srcFeatures[key] * trgFeatures[key])
                elif workset.matchMethod == 'symbols':
                    features[key] += (srcFeatures[key] * trgFeatures[key])
                elif workset.matchMethod == 'treecomp':
                    features[key] += (srcFeatures[key] * trgFeatures[key])
                elif workset.matchMethod == 'treedist':
                    elems1 = recPair[0].trg.split(' ')[0:-1]
                    elems2 = recPair[1].src.split(' ')[0:-1]
                    tree1 = trees.parseSExpression('(' + str.join(' ', elems1) + ')')[0]
                    tree2 = trees.parseSExpression('(' + str.join(' ', elems2) + ')')[0]
                    dist = trees.calcTreeEditDistance(tree1, tree2)
                    numTerms = len(recPair[0].trgTerms)
                    numElems = max(trees.countElements(tree1), trees.countElements(tree2))
                    rate = max(1, 1 + numElems - numTerms - dist) / float(max(1, 1 + numElems - numTerms))
                    features[key] += (rate * srcFeatures[key] * trgFeatures[key])
                elif workset.matchMethod == 'treedistexp':
                    elems1 = recPair[0].trg.split(' ')[0:-1]
                    elems2 = recPair[1].src.split(' ')[0:-1]
                    tree1 = trees.parseSExpression('(' + str.join(' ', elems1) + ')')[0]
                    tree2 = trees.parseSExpression('(' + str.join(' ', elems2) + ')')[0]
                    dist = trees.calcTreeEditDistance(tree1, tree2)
                    rate = math.exp(-dist)
                    features[key] += (rate * srcFeatures[key] * trgFeatures[key])
                else:
                    assert False, 'Invalid Match Method'
        if multi_target:
            # p(trg,pvt|src) ~ p(trg|pvt) * p(pvt|src)
            features['egfp'] =  srcFeatures['egfp'] * trgFeatures['egfp']
            if workset.jointMethod == 'memoryless':
                # p(src|pvt,trg) ~ f(src|pvt)
                features['fgep'] = srcFeatures['fgep']
            else:
                features['fgep'] = 1.0
#            for key in ['egfp', 'fgep']:
#                features.setdefault(key, 0)
#                features[key] += (srcFeatures[key] * trgFeatures[key])
#            # P(trg,pvt|src) = P(trg|pvt,src) * P(pvt|src) ~ P(trg|pvt) * P(pvt|src)
#            features['egfp'] = (srcFeatures['egfp'] * trgFeatures['egfp'])
#            # P(src|pvt,trg) ~ P(src|pvt)
#            features['fgep'] = srcFeatures['fgep']
            for key in ['egfl', 'egfp', 'fgel', 'fgep']:
                features['1'+key] = srcFeatures[key]
    else:
        # multiplying only lexical weights and marginalizing
        for key in ['egfl', 'fgel']:
            features.setdefault(key, 0)
            features[key] += (srcFeatures[key] * trgFeatures[key])
    # using 'p' and 'w' of target
    if 'p' in trgFeatures:
        features['p'] = trgFeatures['p']
    if multi_target:
        if 'w' in trgFeatures:
            features['0w'] = trgFeatures['w']
        if 'w' in srcFeatures:
            features['1w'] = srcFeatures['w']
    else:
        if 'w' in trgFeatures:
            features['w'] = trgFeatures['w']


def updateCounts(recPivot, recPair, method):
    '''update occurrence counts of phrase'''
    counts = recPivot.counts
    features = recPivot.features
    if method == 'countmin':
        #counts.co = max(counts.co, min(recPair[0].counts.co, recPair[1].counts.co))
        #c = recPair[0].counts.co * recPair[1].counts.co
        c = min(recPair[0].counts.cooc, recPair[1].counts.cooc)
        counts.cooc += c
        #counts.co = counts.co + c + 2 * math.sqrt(counts.co * c)
    elif method == 'bidirmin':
        counts1 = recPair[0].counts
        counts2 = recPair[1].counts
        cooc1 = counts1.cooc * counts2.cooc / float(counts2.src)
        cooc2 = counts2.cooc * counts1.cooc / float(counts1.trg)
        counts.cooc += min(cooc1, cooc2)
#        if True:
#        if recPair[0].src.find('Dieu') >= 0:
#            progress.log("%s ||| %s ||| %s ||| (%s %s %s) * (%s %s %s) -> %s %s -> %s\n" % (recPair[0].src, recPair[0].trg, recPair[1].trg, counts1.co, counts1.src, counts1.trg, counts2.co, counts2.src, counts2.trg, co1, co2, min(co1,co2)))
    elif method == 'bidirgmean':
        counts1 = recPair[0].counts
        counts2 = recPair[1].counts
        co1 = counts1.co * counts2.co / float(counts2.src)
        co2 = counts2.co * counts1.co / float(counts1.trg)
        counts.co += math.sqrt(co1*co2)
#        progress.log("%s ||| %s ||| %s ||| (%s %s %s) * (%s %s %s) -> %s %s -> %s\n" % (recPair[0].src, recPair[0].trg, recPair[1].trg, counts1.co, counts1.src, counts1.trg, counts2.co, counts2.src, counts2.trg, co1, co2, math.sqrt(co1*co2)))
    elif method == 'bidirmax':
        counts1 = recPair[0].counts
        counts2 = recPair[1].counts
        co1 = counts1.co * counts2.co / float(counts2.src)
        co2 = counts2.co * counts1.co / float(counts1.trg)
        counts.co += max(co1, co2)
    elif method == 'bidiravr':
        counts1 = recPair[0].counts
        counts2 = recPair[1].counts
        co1 = counts1.cooc * counts2.co / float(counts2.src)
        co2 = counts2.cooc * counts1.co / float(counts1.trg)
        counts.co += (co1 + co2) * 0.5
    elif method == 'prodprob':
        counts.src  = recPair[0].counts.src
        counts.cooc = counts.src * features['egfp']
        counts.trg  = counts.cooc / features['fgep']
    elif method == 'multi':
        c = min(recPair[0].counts.co, recPair[1].counts.co)
        counts.co += c
    else:
        assert False, "Invalid method: %s" % method


def mergeAligns(recPivot, recPair):
    '''merge word alignments'''
#    if recPivot.aligns:
#      return
#    alignSet = set()
    alignMapSrcPvt = recPair[0].alignMap
    alignMapPvtTrg = recPair[1].alignMap
    for srcIndex, pvtIndices in alignMapSrcPvt.items():
        for pvtIndex in pvtIndices:
            for trgIndex in alignMapPvtTrg.get(pvtIndex, []):
              align = '%d-%d' % (srcIndex, trgIndex)
#              alignSet.add(align)
              recPivot.aligns.add(align)
#    recPivot.aligns = sorted(alignSet)


def filterByCountRatioToMax(records, div = 100):
    coMax = 0
    for rec in flattenRecords(records):
        coMax = max(coMax, rec.counts.co)
    if isinstance(records, list):
        newRecords = []
        for rec in records:
            if rec.counts.co >= coMax / float(div):
                newRecords.append( rec )
        records = newRecords
    elif isinstance(records, dict):
        newRecords = {}
        for key, rec in records.items():
            if rec.counts.co >= coMax / float(div):
                newRecords[key] = rec
        records = newRecords
    return records


def calcPhraseTransProbsByCounts(records):
    '''calculate forward phrase trans probs by occurrence counts of the phrases'''
    srcCount = calcSrcCount(records)
    for rec in records.values():
    #for rec in flattenRecords(records):
        counts = rec.counts
        counts.src = srcCount
        if srcCount > 0:
            rec.features['egfp'] = counts.co / float(srcCount)
        else:
            rec.features['egfp'] == 0


def calcPhraseTransProbsOnTable(table_path, savePath, **options):
    '''calculate phrase trans probs on the table in which co-occurrence counts are estimated'''
    method = options.get('method', METHOD)
    RecordClass = options.get('RecordClass', MosesRecord)

    table_file = files.open(table_path, "r")
    saveFile  = files.open(savePath, "w")
    records = {}
    lastSrc = ''
    for line in table_file:
        rec = RecordClass(line)
        key = "%s ||| %s |||" % (rec.src, rec.trg)
        if rec.src != lastSrc and records:
            calcPhraseTransProbsByCounts(records)
            writeRecords(saveFile, records)
            records = {}
        if rec.counts.co > 0:
            records[key] = rec
        lastSrc = rec.src
    if records:
        calcPhraseTransProbsByCounts(records)
        writeRecords(saveFile, records)
    saveFile.close()
    table_file.close()


def calcSrcCount(records):
    '''calculate source phrase occurrence counts by co-occurrence counts'''
    total = 0
    for rec in flattenRecords(records):
        total += rec.counts.co
    return total

def updateWordPairCounts(lexCounts, records):
    '''find word pairs in phrase pairs, and update the counts of word pairs'''
    if len(records) > 0:
        src_symbols = records.values()[0].src_symbols
        if len(src_symbols) == 1:
           for rec in records.values():
               trgSymbols = rec.trgSymbols
               if len(trgSymbols) == 1:
                   lexCounts.addPair(src_symbols[0], trgSymbols[0], rec.counts.co)
        lexCounts.filterNBestBySrc(srcWord = src_symbols[0])

def flattenRecords(records, sort = False):
    '''if records are type of dict, return them as a list'''
    if type(records) == dict:
        if sort:
            recordList = []
            for key in sorted(records.keys()):
              recordList.append(records[key])
            return recordList
        else:
            return records.values()
    elif type(records) == list:
        if sort:
            return sorted(records)
        else:
            return records
    else:
        assert False, "Invalid records"

def pivotRecPairs(rows, workset):
    '''combine the source-pivot and pivot-target records for common pivot phrases

    get the list of record pairs in pivotQueue and put the processed data in outQueue

    if workset.method == "prodprob", estimate trans probs by marginalization
    otherwise, calculate them by estimating co-occurrence counts
    '''
    #lexCounts = lex.PairCounter()

    if len(rows) > 0:
        records = {}
        if workset.multi_target:
            multiRecords = {}
            #jointMethod = workset.jointMethod
        for recPair in rows:
            if workset.matchMethod == 'treecomp':
                # pvt1 and pvt2 must be same
                if recPair[0].trg != recPair[1].src:
                    continue
            elif workset.matchMethod == 'symbols':
                if recPair[0].trgSymbols != recPair[1].src_symbols:
                    continue
            trgKey = recPair[1].trg + ' |||'
            if workset.multi_target:
                #strMultiTrg = intern(recPair[1].trg + ' |COL| ' + recPair[0].trg)
                strMultiTrg = recPair[1].trg + ' |COL| ' + recPair[0].trg
                multiKey = strMultiTrg + ' |||'
            if not trgKey in records:
                # source-target record not yet exists, so making new record
                recPivot = workset.Record()
                recPivot.src = recPair[0].src
                recPivot.trg = recPair[1].trg
                records[trgKey] = recPivot
            recPivot = records[trgKey]
            if workset.multi_target:
                recMulti = workset.Record()
                recMulti.src = recPair[0].src
                recMulti.trg = strMultiTrg
                multiRecords[multiKey] = recMulti
            # estimating updated features
            #updateFeatures(recPivot, recPair, workset.method)
            updateFeatures(recPivot, recPair, workset, False)
            # updating the count of phrase pair
            updateCounts(recPivot, recPair, workset.method)
            #updateCounts(recPivot, recPair, workset)
            # merging the word alignments
            mergeAligns(recPivot, recPair)
            if workset.multi_target:
                #updateFeatures(recMulti, recPair, workset.method, multi_target = True, jointMethod = jointMethod)
                updateFeatures(recMulti, recPair, workset, multi_target = True)
                updateCounts(recMulti, recPair, workset.method)
                mergeAligns(recMulti, recPair)
        # at this time, all the source-target records are determined for given source
        if workset.multi_target:
            # copying the estimated features of source-target records to source-target-pivot records
            for multiKey, recMulti in multiRecords.items():
                trgPair = recMulti.trg.split(' |COL| ')
                recPivot = records[trgPair[0]+' |||']
                for featureKey in ['egfl', 'egfp', 'fgel', 'fgep']:
                    recMulti.features['0'+featureKey] = recPivot.features[featureKey]
                if workset.jointMethod == 'independent':
                    # p(src|pvt,trg) ~ 1 - (1 - p(src|pvt))(1 - p(src|trg))
                    recMulti.features['fgep'] = 1 - (1 - recMulti.features['1fgep']) * (1 - recMulti.features['0fgep'])
#        if workset.method != 'prodprob':
#            # find word pairs in phrase pairs and update the counts of word pairs
#            updateWordPairCounts(lexCounts, records)
#            # filtering n-best records by co-occurrence counts
#            if not NOPREFILTER:
#                if workset.nbest > 0:
#                    if len(records) > workset.nbest:
#                        scores = []
#                        for key, rec in records.items():
#                            scores.append( (rec.counts.co, key) )
#                        scores.sort(reverse = True)
#                        bestRecords = {}
#                        for _, key in scores[:workset.nbest]:
#                            bestRecords[key] = records[key]
#                        records = bestRecords
#            # calculate forward phrase trans probs
#            calcPhraseTransProbsByCounts(records)
        # if threshold is set (non-zero), aborting the records having trans probs under it
        if workset.threshold < 0:
            # aborting records for extremely small trans probs
            ignoring = []
            for key, rec in records.items():
                if rec[0]['fgep'] < workset.threshold and rec[0]['egfp'] < workset.threshold:
                    #ignoring.append(pair)
                    ignoring.append(key)
            for key in ignoring:
                del records[key]
        # if limit number of records is set (non-zero), filter the n-best records by forward trans probs
        if workset.nbest > 0:
            if len(records) > workset.nbest:
                scores = []
                for key, rec in records.items():
                    scores.append( (rec.features['egfp'],key) )
                    #scores.append( (rec.features['fgep'],key) )
                scores.sort(reverse = True)
                bestRecords = {}
                for _, key in scores[:workset.nbest]:
                    bestRecords[key] = records[key]
                records = bestRecords
            if workset.multi_target:
                if len(multiRecords) > workset.nbest:
                    # T1-filtering method
                    bestTrgRecords = {}
                    # first, filtering src-pvt-trg records including n-best src-trg records
                    for multiKey, recMulti in multiRecords.items():
                        for rec in records.values():
                            if multiKey.find(rec.trg + ' |COL|') == 0:
                                bestTrgRecords.setdefault(rec.trg, [])
                                bestTrgRecords[rec.trg].append(recMulti)
                    bestMultiRecords = {}
                    # second, filtering n-best by forward joint trans probs
                    for trgKey, multiList in bestTrgRecords.items():
                        bestMultiRec = None
                        bestForwardJointTransProb = 0
                        for multiRec in multiList:
                            if multiRec.features['egfp'] > bestForwardJointTransProb:
                                bestMultiRec = multiRec
                                bestForwardJointTransProb = multiRec.features['egfp']
                        if bestMultiRec:
                            bestMultiRecords[bestMultiRec.trg] = bestMultiRec
                    # filling n-best records by (s->t,s->t,p)
                    scores = []
                    for multiKey, recMulti in multiRecords.items():
                        scores.append( (recMulti.features['0egfp'],recMulti.features['egfp'],multiKey) )
                    scores.sort(reverse = True)
                    for _, _, multiKey in scores:
                        if len(bestMultiRecords) >= workset.nbest:
                            break
                        else:
                            if multiKey in bestMultiRecords:
                                pass
                            else:
                                bestMultiRecords[multiKey] = multiRecords[multiKey]
                    multiRecords = bestMultiRecords
        if workset.multi_target:
            records = multiRecords
        # putting the records into outQueue, and other process will write them in table file
        if records:
            #workset.pivotCount.add( len(records) )
            for trgKey in sorted(records.keys()):
                rec = records[trgKey]
                writeRecord(rec, workset)
    # exiting from while loop
    # terminate also writeRecords process
    #workset.outQueue.put(None)
    #if workset.method != 'prodprob':
    #    lexCounts.filterNBestByTrg()
    #    lex.saveWordPairCounts(workset.tableLexPath, lexCounts)


def writeRecords(fileObj, records):
#  for rec in flattenRecords(records):
  for rec in flattenRecords(records, sort = True):
      if rec.counts.co > 0:
          fileObj.write( rec.to_str() )
          fileObj.write("\n")


def writeRecord(rec, workset):
    '''write the pivoted records in the queue into the table file'''
    if rec:
        if rec.counts.cooc > 0:
            workset.foutPivot.write( rec.to_str() )
            workset.foutPivot.write( "\n" )
            workset.foutPivot.flush()
            workset.numRecSrcTrg += 1
            workset.setPhrasesSrcTrg.add(rec.src)
            for term in rec.srcTerms:
                workset.setWordsSrcTrg.add(term)

def calcLexWeight(rec, lexCounts, reverse = False):
#    minProb = 10 ** -2
    lexWeight = 1
#    alignMapRev = rec.alignMapRev
    if not reverse:
        minProb  = 1 / float(lexCounts.trgCounts["NULL"])
        # for forward probs, using reversed alignment map
        alignMap = rec.alignMapRev
        srcTerms = rec.srcTerms
        trgTerms = rec.trgTerms
    else:
        minProb = 1 / float(lexCounts.srcCounts["NULL"])
        alignMap = rec.alignMap
        srcTerms = rec.trgTerms
        trgTerms = rec.srcTerms
    minProb = MINPROB
    for trgIndex in range(len(trgTerms)):
        trgTerm = trgTerms[trgIndex]
        if trgIndex in alignMap:
            trgSumProb = 0
            srcIndices = alignMap[trgIndex]
            for srcIndex in srcIndices:
                srcTerm = srcTerms[srcIndex]
                if not reverse:
                    lexProb = lexCounts.calcLexProb(srcTerm, trgTerm)
                else:
                    lexProb = lexCounts.calcLexProbRev(trgTerm, srcTerm)
                trgSumProb += lexProb
            if type(rec) == MosesRecord:
                trgProb = trgSumProb / len(srcIndices)
            else:
                trgProb = trgSumProb / (len(srcIndices) + 1)
#            lexWeight *= (trgSumProb / len(srcIndices))
        else:
          if not reverse:
#              lexWeight *= lexCounts.calcLexProb("NULL", trgTerm)
              trgProb = lexCounts.calcLexProb("NULL", trgTerm)
          else:
#              lexWeight *= lexCounts.calcLexProb(trgTerm, "NULL")
              trgProb = lexCounts.calcLexProb(trgTerm, "NULL")
        lexWeight *= max(trgProb, minProb)
    return lexWeight

def calcLexWeights(table_path, lexCounts, savePath, RecordClass = MosesRecord):
    table_file = files.open(table_path, 'r')
    saveFile  = files.open(savePath, 'w')
    for line in table_file:
        rec = RecordClass(line)
        if rec.trg.find('|COL|') < 0:
            rec.features['egfl'] = calcLexWeight(rec, lexCounts, reverse = False)
            rec.features['fgel'] = calcLexWeight(rec, lexCounts, reverse = True)
            saveFile.write( rec.to_str() )
        else:
            rec.features['0egfl'] = calcLexWeight(rec, lexCounts, reverse = False)
            rec.features['0fgel'] = calcLexWeight(rec, lexCounts, reverse = True)
            saveFile.write( rec.to_str() )
    saveFile.close()
    table_file.close()


def pivot(table1, table2, savefile="phrase-table.gz", workdir=".", **options):
    '''find pair of source-pivot and pivot-target records for common pivot phrase'''
    try:
        # initialize the options
        RecordClass = options.get('RecordClass', MosesRecord)
        prefix = options.get('prefix', 'phrase')
        threshold = options.get('threshold', THRESHOLD)
        alignLexPath   = options.get('alignlex', None)
        nbest     = options.get('nbest', NBEST)
        method    = options.get('method', METHOD)
        lexMethod = options.get('lexmethod', LEX_METHOD)
        jointMethod = options.get('jointmethod', JOINT_METHOD)
        matchMethod = options.get('matchmethod', MATCH_METHOD)
        numNulls  = options.get('nulls', NULLS)
        multi_target = options.get('multitarget', False)
        logFile = options.get('log', '')
        showProgress = options.get('progress', True)

        if lexMethod not in ('prodweight', 'table'):
            if alignLexPath == None:
                logging.debug(lexMethod)
                assert False, "aligned lexfile is not given"

        if matchMethod == 'hiero':
            key_type = 'src_hiero'
        else:
            key_type = 'src_symbols'
        logging.log("loading: %s" % table1)
        #tableSrcPvt = Table(table1, RecordClass, key_type=key_type, showProgress=showProgress)
        tableSrcPvt = Table(table1, RecordClass, showProgress=showProgress)
        logging.log("loading: %s" % table2)
        tablePvtTrg = Table(table2, RecordClass, key_type=key_type, showProgress=showProgress)

        workOptions = {}
        workOptions['RecordClass'] = RecordClass
        workOptions['prefix'] = prefix
        workOptions['multi_target'] = multi_target
        workOptions['jointMethod'] = jointMethod
        workOptions['matchMethod'] = matchMethod
        workset = WorkSet(savefile, workdir, method, **workOptions)
        workset.threshold = threshold
        workset.nbest = nbest

        rows = []
        lastSrc = ''
        logging.log("beginning pivot\n")
        #for recSrcPvt in progress.view(tableSrcPvt.find(''),maxCount=len(tableSrcPvt)):
        for recSrcPvt in progress.view(tableSrcPvt.find(''),'processing',max_count=len(tableSrcPvt)):
            src = recSrcPvt.src
            workset.numRecSrcPvt += 1
            workset.setPhrasesSrcPvt.add(recSrcPvt.src)
            for term in recSrcPvt.srcTerms:
                workset.setWordsSrcPvt.add(term)
            if src != lastSrc:
                if len(rows) > 0:
                    pivotRecPairs(rows, workset)
                    rows = []
            #pvtKey = str.join(' ', recSrcPvt.trgSymbols)
            #if key_type == 'src_hiero':
            #    pvtKey = str.join(' ', RecordClass.getSymbols(recSrcPvt.trg,hiero=True))
            #elif key_type == 'src_symbols':
            #    pvtKey = str.join(' ', RecordClass.getSymbols(recSrcPvt.trg,hiero=False))
            #logging.log("pvt key: %s" % (pvtKey,))
            #for recPvtTrg in tablePvtTrg.find(pvtKey):
            #for recPvtTrg in tablePvtTrg.find_src(pvtKey):
            for recPvtTrg in tablePvtTrg.find_src(recSrcPvt.trg):
                workset.numRecPvtTrg += 1
                workset.setPhrasesPvtTrg.add(recPvtTrg.src)
                for term in recPvtTrg.srcTerms:
                    workset.setWordsPvtTrg.add(term)
                rows.append( (recSrcPvt, recPvtTrg) )
            lastSrc = src
        if len(rows) > 0:
            pivotRecPairs(rows, workset)
            rows = []
        if logFile:
            with open(logFile, 'w') as fobj:
                fobj.write("%s = %s\n" % ('numRecSrcPvt', workset.numRecSrcPvt))
                fobj.write("%s = %s\n" % ('uniqPhrasesSrcPvt', len(workset.setPhrasesSrcPvt)))
                fobj.write("%s = %s\n" % ('uniqWordsSrcPvt', len(workset.setWordsSrcPvt)))
                fobj.write("--\n")
                fobj.write("%s = %s\n" % ('numRecPvtTrg', workset.numRecPvtTrg))
                fobj.write("%s = %s\n" % ('uniqPhrasesPvtTrg', len(workset.setPhrasesPvtTrg)))
                fobj.write("%s = %s\n" % ('uniqWordsPvtTrg', len(workset.setWordsPvtTrg)))
                fobj.write("--\n")
                fobj.write("%s = %s\n" % ('numRecSrcTrg', workset.numRecSrcTrg))
                fobj.write("%s = %s\n" % ('uniqPhrasesSrcTrg', len(workset.setPhrasesSrcTrg)))
                fobj.write("%s = %s\n" % ('uniqWordsSrcTrg', len(workset.setWordsSrcTrg)))
                fobj.write("--\n")
                if len(workset.setWordsSrcPvt) != len(workset.setWordsSrcTrg):
                    fobj.write("Lost Words:\n")
                    for word in iter(workset.setWordsSrcPvt - workset.setWordsSrcTrg):
                        fobj.write("\t* \"%s\"\n" % word)
        workset.close()
#        # loading necessary word pair count files
#        if lexMethod != 'prodweight':
#            if lexMethod.find('table') >= 0:
#                if lexMethod.find('+') >= 0:
#                    progress.log("combining lex counts into: %s\n" % (workset.combinedLexPath))
#                    combine_lex.combine_lex(alignLexPath, workset.tableLexPath, workset.combinedLexPath)
#                    progress.log("loading combined word trans probabilities\n")
#                    lexCounts = lex.loadWordPairCounts(workset.combinedLexPath)
#                else:
#                    progress.log("loading table lex: %s\n", workset.tableLexPath)
#                    lexCounts = lex.loadWordPairCounts(workset.tableLexPath)
#                    lexCounts.srcCounts["NULL"] = numNulls
#                    lexCounts.trgCounts["NULL"] = numNulls
#            else:
#                progress.log("loading aligned lex: %s\n" % alignLexPath)
#                lexCounts = lex.loadWordPairCounts(alignLexPath)
##        if workset.method == 'countmin':
#        if method in ['countmin', 'bidirmin', 'bidirgmean', 'bidirmax', 'bidiravr']:
##            # 単語単位の翻訳確率をロードする
##            #progress.log("loading word trans probabilities\n")
##            #lexCounts = lex.loadWordPairCounts(lexPath)
##            progress.log("combining lex counts into: %s\n" % (workset.combinedLexPath))
##            combine_lex.combine_lex(lexPath, workset.tableLexPath, workset.combinedLexPath)
##            progress.log("loading combined word trans probabilities\n")
##            lexCounts = lex.loadWordPairCounts(workset.combinedLexPath)
#            # reversing the table
#            progress.log("reversing %s table into: %s\n" % (prefix, workset.revPath) )
#            reverseTable(workset.pivotPath, workset.revPath, RecordClass)
#            progress.log("reversed %s table\n" % (prefix))
#            # calculating backward phrase trans probs for reversed table
#            progress.log("calculating reversed phrase trans probs into: %s\n" % (workset.trgCountPath))
#            calcPhraseTransProbsOnTable(workset.revPath, workset.trgCountPath, nbest = workset.nbest, RecordClass = RecordClass)
#            progress.log("calculated reversed phrase trans probs\n")
#            # reverseing the reversed table
##            progress.log("reversing %s table into: %s\n" % (prefix,workset.revTrgCountPath))
#            progress.log("reversing %s table into: %s\n" % (prefix,workset.countPath))
##            reverseTable(workset.trgCountPath, workset.revTrgCountPath, RecordClass)
#            reverseTable(workset.trgCountPath, workset.countPath, RecordClass)
#            progress.log("reversed %s table\n" % (prefix))
#            # calculating the forward trans probs
##            progress.log("calculating phrase trans probs into: %s\n" % (workset.countPath))
##            calcPhraseTransProbsOnTable(workset.revTrgCountPath, workset.countPath, nbest = 0, RecordClass = RecordClass)
##            progress.log("calculated phrase trans probs\n")
#            if lexMethod != 'prodweight':
#                # calculating lexical weights
#                progress.log("calculating lex weights into: %s\n" % workset.savePath)
#                calcLexWeights(workset.countPath, lexCounts, workset.savePath, RecordClass)
#                progress.log("calculated lex weights\n")
#            else:
#                progress.log("gzipping into: %s\n" % workset.savePath)
#                files.autoCat(workset.countPath, workset.savePath)
##        elif method == 'prodprob':
#        elif method.find('prodprob') >= 0:
#            if lexMethod != 'prodweight':
#                # calculating lexical weights
#                progress.log("calculating lex weights into: %s\n" % workset.savePath)
#                calcLexWeights(workset.pivotPath, lexCounts, workset.savePath, RecordClass)
#                progress.log("calculated lex weights\n")
#            else:
#                progress.log("gzipping into: %s\n" % workset.savePath)
#                files.autoCat(workset.pivotPath, workset.savePath)
##        elif method == 'multi':
##                progress.log("gzipping into: %s\n" % workset.savePath)
##                files.autoCat(workset.pivotPath, workset.savePath)
#        else:
#            assert False, "Invalid method: %s" % method
    except KeyboardInterrupt:
        # catching exception, finish all the workset processes
        print('')
        print('Caught KeyboardInterrupt, terminating all the worker processes')
        workset.close()
        sys.exit(1)

def main():
    parser = argparse.ArgumentParser(description = 'load 2 rule tables and pivot into one travatar rule table')
    parser.add_argument('table1', help = 'rule table 1')
    parser.add_argument('table2', help = 'rule table 2')
    parser.add_argument('savefile', help = 'path for saving travatar rule table file')
    parser.add_argument('--threshold', help = 'threshold for ignoring the phrase translation probability (real number)', type=float, default=THRESHOLD)
    parser.add_argument('--nbest', help = 'best n scores for rule pair filtering (default = 20)', type=int, default=NBEST)
    parser.add_argument('--method', help = 'triangulation method', choices=method_list, default=METHOD)
    parser.add_argument('--lexmethod', help = 'lexical triangulation method', choices=lex_method_list, default=LEX_METHOD)
    parser.add_argument('--jointmethod', help = 'method to estimate joint trans probs', choices=joint_method_list, default=JOINT_METHOD)
    parser.add_argument('--matchmethod', help = 'matching method', choices=match_method_list, default=MATCH_METHOD)
    parser.add_argument('--workdir', help = 'working directory', default='.')
    parser.add_argument('--alignlex', help = 'word pair counts file', default=None)
    parser.add_argument('--nulls', help = 'number of NULLs (lines) for table lex', type = int, default=NULLS)
    parser.add_argument('--noprefilter', help = 'No pre-filtering', type = bool, default=False)
    parser.add_argument('--multitarget', help = 'enabling multi target model', action='store_true')
    parser.add_argument('--progress', '-p', help = 'show progress', type = bool, default=True)
    parser.add_argument('--log', help = 'log file (optional)', type = str, default='')
    args = vars(parser.parse_args())

    args['RecordClass'] = TravatarRecord
    args['prefix'] = 'rule'
    #base.pivot(**args)
    pivot(**args)

if __name__ == '__main__':
    main()




# distutils: language=c++
# -*- coding: utf-8 -*-

'''Utility functions handling colors'''

# C++ setting
from libcpp cimport bool

from lpu.__system__ import logging

logger = logging.getLogger(__name__)

COLOR_MAP = {
    'clear' : '\033[0m',
    'black' : '\033[30m',
    'red'   : '\033[31m',
    'green' : '\033[32m',
    'yellow': '\033[33m',
    'blue'  : '\033[34m',
    'purple': '\033[35m',
    'cyan'  : '\033[36m',
    'white' : '\033[37m'
}

ctypedef fused strtype:
    str
    bytes

#def put_color(content, color=None, eachline=True):
#cpdef str put_color(content, str color=None, bool eachline=True):
cpdef str put_color(str content, str color=None, bool eachline=True):
#cpdef str put_color(strtype content, strtype color=None, bool eachline=True):
    #print("color name: {}".format(color))
    code = COLOR_MAP.get(color, None)
    #code = colors[color]
    #print("code: {}".format(repr(code)))
    if code:
        if eachline:
            lines = str(content).split('\n')
            lines = ['{}{}{}'.format(code, line.rstrip(), COLOR_MAP['clear']) for line in lines]
            return str.join('\n', lines)
        else:
            #return "%s%s%s" % (code, content, COLOR_MAP['clear'])
            return '{}{}{}'.format(code, content, COLOR_MAP['clear'])
    else:
        return content




#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''function converting rule table'''

# Standard libraries
import argparse
import subprocess

# Local libraries
from lpu.common import files
from lpu.common.progress import view
from lpu.smt.trans_models import records

CONVERT_OPTIONS=["scfg", "hiero", "tag"]

def convertTravatarExtract(srcFile, saveFile, sync=None, flatten=None, reverse=False, no_unary=False, progress=False):
    if type(srcFile) == str:
        #srcFile = files.open(srcFile)
        srcFile = files.open(srcFile, 'rt')
    if type(saveFile) == str:
        saveFile = files.open(saveFile, 'wt')
    if progress:
        srcFile = view(srcFile)
    for line in srcFile:
        fields = line.strip().split('|||')
        fields = [f.strip() for f in fields]
        if len(fields) >= 4:
            srcSymbols = fields[0].strip().split(' ')
            trgSymbols = fields[1].strip().split(' ')
            count = fields[2].strip()
            if sync:
                srcSymbols, trgSymbols = records.syncTags(srcSymbols, trgSymbols, sync)
                trgSymbols, srcSymbols = records.syncTags(trgSymbols, srcSymbols, sync)
            if flatten:
                srcSymbols = records.getFlattenSymbols(srcSymbols, flatten)
                trgSymbols = records.getFlattenSymbols(trgSymbols, flatten)
            if reverse:
                srcSymbols, trgSymbols = trgSymbols, srcSymbols
                aligns = records.getAlignSet(fields[3])
                aligns = records.getRevAlignSet(aligns)
                fields[3] = str.join(' ', sorted(aligns))
            if no_unary:
                #if len(records.getTravatarTerms(srcSymbols)) == 0:
                if len(records.TravatarRecord.getTerms(srcSymbols)) == 0:
                    # no terminals
                    continue
#                if len(srcSymbols) == 3:
#                    tag = srcSymbols[2]
#                    if srcSymbols[0] == "x0:"+tag:
#                        # src is unary cycle
#                        continue
            fields[0] = str.join(' ', srcSymbols)
            fields[1] = str.join(' ', trgSymbols)
            fields[0], fields[1] = records.TravatarRecord.fixOrderStrings(fields[0], fields[1])
            saveFile.write(str.join(' ||| ', fields)+"\n")
    saveFile.close()

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('src_table', help = 'source rule table')
    parser.add_argument('save_table', help = 'save path')
    parser.add_argument('--sync', choices=CONVERT_OPTIONS, help='transfer the phrase tags from subtrees')
    parser.add_argument('--flatten', choices=CONVERT_OPTIONS, help='flatten the trees into symbols string in specified style')
    parser.add_argument('--reverse', action='store_true', help='reverse the trasnlation direction of records')
    parser.add_argument('--no-unary', dest='no_unary', action='store_true', help='avoid unary cycle')
    parser.add_argument('--progress', '-p', action='store_true', help='show progress')
    args = parser.parse_args()
    #args = vars(parser.parse_args())
    #print(args)
    convertTravatarExtract(args.src_table, args.save_table, args.sync, args.flatten, args.reverse, args.no_unary, args.progress)

if __name__ == '__main__':
    main()




#!/usr/bin/env python

import argparse
import os
import sys

def printAbsPathList(filepaths):
    print(str.join(' ', map(os.path.abspath, filepaths)))

def cmdAbsPath(args):
    parser = argparse.ArgumentParser(description='Get absolute paths of given files or directories')
    parser.add_argument('filepaths', metavar="filepath", nargs="+", type=str, help='path of file to get the absolute paths')
    parsed = parser.parse_args(args)
    printAbsPathList(parsed.filepaths)

def main():
    cmdAbsPath(sys.argv[1:])

if __name__ == '__main__':
    main()

